/*
 * User Mode Sandbox for Grid
 * Copyright (C) 2004 Tangui Morlier
 *     Grand-Large (INRIA Futurs)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#include "syscall_names.h"
#include "sandbox.h"
#include <errno.h>
#ifdef  NR_syscalls
#define NBDEFSYSCALLS NR_syscalls
#else//NR_syscalls
#ifdef __NR_security
#define NBDEFSYSCALLS 253
#else
#define NBDEFSYSCALLS 222
#endif
#endif//NR_syscalls


int  nb_syscall2filter = 0;

/**
 * Block all precesses execution when a syscall is filtered
 */

void barrier_on(struct pid_info * my_pid_info) {
  int i, status;
  pid_t pid;
#ifdef VERBOSE_PLUS
  fprintf(stderr, "Barrier ON\n");
#endif //VERBOSE_PLUS
  my_pid_info->barrier = 1;
  for(i = 0 ; i<MAX_PID_INFO ; i++){
    if (pid_infos[i].pi.pid == 0)
      break;
    if (pid_infos[i].pi.pid != my_pid_info->pi.pid && pid_infos[i].filterSyscall){
      kill(pid_infos[i].pi.pid, SIGBARRIER);
      wait4(pid_infos[i].pi.pid, &status, 0, NULL), WSTOPSIG(status);
      
      //  We may have received an other signal than the SIGBARRIER
      //we've just sent. In this case this signal will be forwarded 
      //to the sandboxed process when it will be woken up or 
      pid_infos[i].sig2forward = 0;
      pid_infos[i].filterSyscall = 0;
      if(WSTOPSIG(status) != SIGBARRIER)
	if(WSTOPSIG(status) == SIGTRAP) {
	  pid_infos[i].filterSyscall = 1;
	  nb_syscall2filter++;
	  pid_infos[i].sig2forward = status;
	}else{
	  pid_infos[i].sig2forward = WSTOPSIG(status);
	}
    }
  }
}

/**
 * Unblock the processes
 */
void barrier_off(struct pid_info * my_pid_info) {
#ifdef VERBOSE_PLUS
  fprintf(stderr, "Barrier OFF\n");
#endif //VERBOSE_PLUS
  int i;
  my_pid_info->barrier = 0;
  for(i = 0 ; i<MAX_PID_INFO ; i++){
    if (pid_infos[i].pi.pid == 0)
      break;
    if (pid_infos[i].pi.pid != my_pid_info->pi.pid && !pid_infos[i].filterSyscall){

      ptrace(PTRACE_SYSCALL, pid_infos[i].pi.pid, 0, pid_infos[i].sig2forward);
    }
  }
}


/**
 * Remove a PID for the list of children processes queue
 */
int removePID(pid_t pid){
  int i, index=0;
  ptrace(PTRACE_KILL, pid, 0, 0);
  for(i = 0 ; i < MAX_PID_INFO ; i++) {
    if(pid_infos[i].pi.pid == 0)
      break;
    if(pid_infos[i].pi.pid == pid){
      index = i;
      break;
    }
  }


  for(i++; i < MAX_PID_INFO ; i++) {
    if(pid_infos[i].pi.pid == 0){
      if(i-1 != index)
	memcpy(&pid_infos[index], &pid_infos[i-1], sizeof(struct pid_info));
      bzero(&pid_infos[i-1], sizeof(struct pid_info));
      break;
    }
  }
  if(pid_infos[0].pi.pid == 0)
    return 0;
  return 1;
}

/**
 * Add a pid in the children processes queue if unknown
 * Return the pid_info structure corresponding to this pid
 */
struct pid_info * addPID(pid_t pid) {
  int i;
  struct pid_info * my_pid_info;

  for(i = 0 ; i < MAX_PID_INFO ; i++) {
    if(pid_infos[i].pi.pid == pid) {
      my_pid_info = &pid_infos[i];
      break;
    }
    if(pid_infos[i].pi.pid == 0) {

      pid_infos[i].pi.pid = pid;
      pid_infos[i].pi.in_or_out = 0;
      my_pid_info = &pid_infos[i];
      break;
    }
  }
  if(i == MAX_PID_INFO){
    for(i = 0 ; i < MAX_PID_INFO ; i++)
      ptrace(PTRACE_KILL, pid_infos[i].pi.pid, 0, 0);
    exit(-20);
  }
}

/**
 * wait for the next syscall
 **/
pid_t wait_ptrace( struct pid_info * my_pid_info, int * status, int * sig ){
  int i;
  pid_t pid;
  //Wake up the last sandboxed process
  ptrace(PTRACE_SYSCALL, my_pid_info->pi.pid, 0, sig);
  //If we are in a barrier wait only for the last sandboxed process
  if(my_pid_info->pi.in_or_out && my_pid_info->barrier) {
    pid = wait4(my_pid_info->pi.pid, status, __WALL, NULL);
    if (pid == -1) {
      fprintf(stderr, "An error occured on Wait: %s\n", strerror(errno));
    }
    return pid;
  }
  // If during the barrier, we received a ptrace signal instead of
  //an internal one, we simulate the reception of this signal
  if(nb_syscall2filter) {
    for(i = 0 ; i < MAX_PID_INFO ; i++){
      if(pid_infos[i].pi.pid == 0)
	break;
      if (pid_infos[i].filterSyscall) {
	nb_syscall2filter--;
	pid_infos[i].filterSyscall;
	*status = pid_infos[i].sig2forward;
	return pid_infos[i].pi.pid;
      }
    }
  }

  pid = wait4(-1, status, __WALL, NULL);
  if (pid == -1) {
    fprintf(stderr, "An error occured on Wait: %s\n", strerror(errno));
  }
  return pid;

}

/**
 * Realize the sandbox activities
 */
int sandboxIt(char* process_name) {

  int i, status, received_sig;
  pid_t pid;
  rule_function * rule;
  struct pid_info * my_pid_info;

#ifdef VERBOSE
  int multi = 0;
  fprintf(stderr, "SandBox !!\n");
#endif

  bzero(&pid_infos, sizeof(struct pid_info)*MAX_PID_INFO);
  
  //The first signal sent correspond to the return exeve syscall
  //We ignore it to have a easier in_or_out status gestion
  pid = wait4(-1, &status, __WALL, NULL);


  /**
   * Begin to sandbox the application
   */
  ptrace(PTRACE_SYSCALL, pid, NULL, NULL);
  pid = wait4(-1, &status, __WALL, NULL);
  
  while (1) {

    if (!WIFSTOPPED(status) && !WIFEXITED(status) && !WIFSIGNALED(status)) 
      continue;

    //Register if unknown the pid and get infos
    my_pid_info = addPID(pid);
    received_sig = WSTOPSIG(status);
    //If the signal is not a syscall one
    if(received_sig != SIGTRAP) { 
      //It may be a child end
      if(received_sig == SIGKILL || WIFEXITED(status) || WIFSIGNALED(status)){
	if(!removePID(pid))
	  break;
	/*	else
		received_sig = (int) NULL;*/
      }

      //If at last barrier, we forwarded an other signal than SIGCHLD,
      //We are receiving a SIGCHLD signal which is not intended to the process
      //This SIGCHLD is then ignored
      if(received_sig == SIGBARRIER && my_pid_info->sig2forward){
	my_pid_info->sig2forward = 0;
	received_sig = 0;
      }
      if(received_sig == SIGCONT) {
	received_sig = 0;
      }


      //Send the signal to the child and wait for the next ptrace signal
      pid = wait_ptrace(my_pid_info, &status, (void *)received_sig);
      continue;
    }
    
    //Get regs info to analyse the syscall
    ptrace(PTRACE_GETREGS, my_pid_info->pi.pid, NULL, &(my_pid_info->pi.regs));
    if(my_pid_info->pi.regs.orig_eax != my_pid_info->last_syscall)
      my_pid_info->pi.in_or_out = 0;
    
#ifdef VERBOSE_PLUS
      fprintf(stderr, "[%d %s] %s called (i|o ? %d) (sc #%d)\n", 
	my_pid_info->pi.pid, process_name, syscall_names[my_pid_info->pi.regs.orig_eax], my_pid_info->pi.in_or_out, my_pid_info->pi.regs.orig_eax);
#endif //VERBOSE_PLUS

      //If a rule function is known for this syscall
      if(my_pid_info->pi.regs.orig_eax > 0 && my_pid_info->pi.regs.orig_eax < NBDEFSYSCALLS) {
	if (syscall_rules[my_pid_info->pi.regs.orig_eax].rules[my_pid_info->pi.in_or_out]!=NULL){
	  //If we may filter this syscall, we use a barrier
	  if (!my_pid_info->pi.in_or_out) {
	    barrier_on(my_pid_info);
	  }
	  
	  //If this syscall is not allowed, 
	  //we stop the execution of all child processes
	  my_pid_info->msg[0] = '\0';
	  if(!syscall_rules[my_pid_info->pi.regs.orig_eax].rules[my_pid_info->pi.in_or_out]
	     (&(my_pid_info->pi), my_pid_info->str, my_pid_info->msg)){
	    for(i = 0 ; i < MAX_PID_INFO ; i++)
	      if (pid_infos[i].pi.pid == 0)
		break;
	      else
		ptrace(PTRACE_KILL, pid_infos[i].pi.pid, 0, 0);
	    //Print the message returned by the rule function
	    fprintf(stderr, "[%d] %s stoped while calling '%s':\n\t%s\n", 
		    my_pid_info->pi.pid, process_name, 
		    syscall_names[my_pid_info->pi.regs.orig_eax], my_pid_info->msg );
	    exit(-20);
	  }
#ifdef VERBOSE
	  else{
	    fprintf(stderr, "[%d] Sandbox allows the execution of %s\n", 
		    my_pid_info->pi.pid, syscall_names[my_pid_info->pi.regs.orig_eax]);
	  }
#endif //VERBOSE
	}
#ifdef VERBOSE_PLUS
	else{
	  fprintf(stderr, "[%d] %s (%s) does not have sandbox rule\n", 
		  my_pid_info->pi.pid, 
		  syscall_names[my_pid_info->pi.regs.orig_eax],
		  (my_pid_info->pi.in_or_out) ? "out":"in");
	}
#endif //VERBOSE_PLUS
      }else{
	for(i = 0 ; i < MAX_PID_INFO ; i++)
	  if (pid_infos[i].pi.pid == 0)
	    break;
	  else
	    ptrace(PTRACE_KILL, pid_infos[i].pi.pid, 0, 0);
	//Print the message returned by the rule function
	fprintf(stderr, "[%d] %s stoped while calling '%s' : \n\t%s\n", 
		my_pid_info->pi.pid, process_name, 
		syscall_names[my_pid_info->pi.regs.orig_eax], my_pid_info->msg );
	exit(-20);
      }
      //Desactive the barrier on syscall return
      if(my_pid_info->pi.in_or_out && my_pid_info->barrier){
	barrier_off(my_pid_info);
      }
      
      if(my_pid_info->pi.regs.orig_eax == SYS_exit && my_pid_info->pi.in_or_out) {
	if(!removePID(pid))
	  break;
      }
    
      if(my_pid_info->pi.regs.orig_eax == SYS_execve 
	 && !my_pid_info->pi.regs.ecx) {
	my_pid_info->pi.in_or_out = 1;
      }
      //Monitor if we get in or out of the syscall
      my_pid_info->pi.in_or_out ^= 1;
      my_pid_info->last_syscall = my_pid_info->pi.regs.orig_eax;
      pid = wait_ptrace(my_pid_info, &status, NULL);
  }
  
  for(i = 0 ; i < MAX_PID_INFO ; i++)
    ptrace(PTRACE_KILL, pid_infos[i].pi.pid, 0, 0);

  if(WIFEXITED(status))
    return WEXITSTATUS(status);

  return 0;
}
