/*
 * User Mode Sandbox for Grid
 * Copyright (C) 2004 Tangui Morlier
 * Copyright (C) 2004 Grand-Large (INRIA Futurs)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#include "sandboxRules.h"
#include <sys/socket.h>
#include <linux/net.h>
int is_first_execve = 1;
int is_first_uname = 1;

/**
 * A basic log function
 */
int sandbox_log(struct ptrace_info_struct * pi, char* ptr, char* msg) {
  fprintf(stderr, "[%d] --> %s called\n", pi->pid, syscall_names[pi->regs.orig_eax]);
  return 1;
}

/**
 * Deny a syscall
 */
int sandbox_deny(struct ptrace_info_struct * pi, char* ptr, char* msg) {

  int eax_orig;

  snprintf(msg, MAX_MSG_SIZE, "%s is a denied system call", syscall_names[pi->regs.orig_eax]);

  /* If fork like functions are denied, we have to stop the execution 
     of its child 
     We then rewrite the syscall to redirect it to the exit function 
     instead of fork
  */
  if(pi->regs.orig_eax == SYS_fork || pi->regs.orig_eax == SYS_vfork || 
     pi->regs.orig_eax == SYS_clone) {
    eax_orig = pi->regs.orig_eax;
    pi->regs.orig_eax = SYS_exit;
    pi->regs.ebx = 0;
    ptrace(PTRACE_SETREGS, pi->pid, NULL, &(pi->regs));
    pi->regs.orig_eax = eax_orig;
  }

  return 0;
}


int socket_filter(struct ptrace_info_struct * pi, char* ptr, char* msg) {
  /*
   * Authorized only the unix and local sockets
   */

  int first;
  if(pi->regs.ebx == SYS_SOCKET) {
    first = decodeWord(pi->pid, pi->regs.ecx);
    if (first==PF_UNIX || first==PF_LOCAL) 
      return 1;
    snprintf(msg, MAX_MSG_SIZE, "Sorry, only local and unix sockets are allowed");
    return 0;
  }
  return 1;
}

int which(const char* file, char* buffer) {
  char *end, *first;
  char real[PATH_MAX];
  int found = 0;

  if(file[0] == '.' || file[0] == '/')
    return 0;
  first = getenv("PATH");
  found = 0;
  while(1) {
    end = strchr(first, ':');
    if(end != NULL)
      end[0] = '\0';
    strncpy(buffer, first, PATH_MAX);
    strncat(buffer, "/", PATH_MAX);
    strncat(buffer, file, PATH_MAX);
    if(realpath(buffer, real) != NULL) {
      found = 1;
      break;
    }
    if(end == NULL)
      break;
    end[0] = ':';
    first = end +1;
  }
  return found;
}

int is_ptracable_bin(const char * file) {
  char buffer[5];
  int fd;
  fd = open(file, O_RDONLY);
  read(fd, &buffer, 4);
  close(fd);
  buffer[4] = '\0';
  
  if(buffer[0] == '#')
    return 1;
  if(strcmp(&buffer[1], "ELF") == 0)
    return 1;
  return 0;
}

/**
 * Initializate sandbox rules
 */
void generateSandboxRules() {
  bzero(&syscall_rules, sizeof(struct syscall_rules_struct)*NB_SYSCALLS);


  /**
   * When fork, we follow the execution of the child
   *    With multiproc/thread, all processes are suspended when 
   *   a syscall is filtered
   *  see fork_functions.h for more info
   */
  syscall_rules[SYS_fork].rules[1] = &fork_attitude;
  syscall_rules[SYS_vfork].rules[1] = &fork_attitude;
  syscall_rules[SYS_clone].rules[1] = &fork_attitude;


  /**
   * All dangerous syscalls are denied
   *  Their call provoques the application destruction
   */

  syscall_rules[SYS_idle].rules[0] = &sandbox_deny;
  syscall_rules[SYS_gtty].rules[0] = &sandbox_deny;
  syscall_rules[SYS_lock].rules[0] = &sandbox_deny;
  syscall_rules[SYS_mknod].rules[0] = &sandbox_deny;
  syscall_rules[SYS_mount].rules[0] = &sandbox_deny;
  syscall_rules[SYS_oldstat].rules[0] = &sandbox_deny;
  syscall_rules[SYS_prof].rules[0] = &sandbox_deny;
  syscall_rules[SYS_ptrace].rules[0] = &sandbox_deny;
  syscall_rules[SYS_reboot].rules[0] = &sandbox_deny;
  syscall_rules[SYS_sethostname].rules[0] = &sandbox_deny;
  syscall_rules[SYS_settimeofday].rules[0] = &sandbox_deny;
#ifdef SYS_statfs64
  syscall_rules[SYS_statfs64].rules[0] = &sandbox_deny;
#endif
  syscall_rules[SYS_statfs].rules[0] = &sandbox_deny;
  syscall_rules[SYS_stty].rules[0] = &sandbox_deny;
  syscall_rules[SYS_swapon].rules[0] = &sandbox_deny;
  syscall_rules[SYS_swapoff].rules[0] = &sandbox_deny;
  syscall_rules[SYS_umount].rules[0] = &sandbox_deny;
  syscall_rules[SYS_umount2].rules[0] = &sandbox_deny;
  syscall_rules[SYS_ustat].rules[0] = &sandbox_deny;
  syscall_rules[SYS_adjtimex].rules[0] = &sandbox_deny;
  syscall_rules[SYS_bdflush].rules[0] = &sandbox_deny;
  syscall_rules[SYS_capget].rules[0] = &sandbox_deny;
  syscall_rules[SYS_capset].rules[0] = &sandbox_deny;
#ifdef  SYS_clock_settime
  syscall_rules[SYS_clock_settime].rules[0] = &sandbox_deny;
#endif
  syscall_rules[SYS_create_module].rules[0] = &sandbox_deny;
  syscall_rules[SYS_delete_module].rules[0] = &sandbox_deny;
  syscall_rules[SYS_init_module].rules[0] = &sandbox_deny;
  syscall_rules[SYS_query_module].rules[0] = &sandbox_deny;
  syscall_rules[SYS_get_kernel_syms].rules[0] = &sandbox_deny;
  syscall_rules[SYS_modify_ldt].rules[0] = &sandbox_deny;
  //  syscall_rules[SYS_uname].rules[0] = &sandbox_deny; //<----------
  syscall_rules[SYS_olduname].rules[0] = &sandbox_deny;
  syscall_rules[SYS_oldolduname].rules[0] = &sandbox_deny;
  syscall_rules[SYS_personality].rules[0] = &sandbox_deny;
  syscall_rules[SYS_pivot_root].rules[0] = &sandbox_deny;
  syscall_rules[SYS_quotactl].rules[0] = &sandbox_deny;
#ifdef SYS_sched_getaffinity
  syscall_rules[SYS_sched_getaffinity].rules[0] = &sandbox_deny;
#endif
  syscall_rules[SYS_sched_getparam].rules[0] = &sandbox_deny;
#ifdef SYS_sched_setaffinity
  syscall_rules[SYS_sched_setaffinity].rules[0] = &sandbox_deny;
#endif
  syscall_rules[SYS_sched_setparam].rules[0] = &sandbox_deny;
  syscall_rules[SYS_sched_setscheduler].rules[0] = &sandbox_deny;
  //  syscall_rules[SYS_set_tid_address].rules[0] = &sandbox_deny;
  syscall_rules[SYS_setdomainname].rules[0] = &sandbox_deny;
  //  syscall_rules[SYS_setrlimit].rules[0] = &sandbox_deny; //<----------
  syscall_rules[SYS_stime].rules[0] = &sandbox_deny;
  syscall_rules[SYS_sysfs].rules[0] = &sandbox_deny;
  syscall_rules[SYS_sysinfo].rules[0] = &sandbox_deny;
  syscall_rules[SYS_syslog].rules[0] = &sandbox_deny;
  syscall_rules[SYS_vhangup].rules[0] = &sandbox_deny;
  syscall_rules[SYS_vm86old].rules[0] = &sandbox_deny;
  syscall_rules[SYS_vm86].rules[0] = &sandbox_deny;
#ifdef SYS_vserver
  syscall_rules[SYS_vserver].rules[0] = &sandbox_deny;
#endif
  syscall_rules[SYS_getpmsg].rules[0] = &sandbox_deny;
  syscall_rules[SYS_afs_syscall].rules[0] = &sandbox_deny;
  syscall_rules[SYS_putpmsg].rules[0] = &sandbox_deny;
  syscall_rules[SYS_prctl].rules[0] = &sandbox_deny;
  syscall_rules[SYS_ioperm].rules[0] = &sandbox_deny;
  syscall_rules[SYS_iopl].rules[0] = &sandbox_deny;
  syscall_rules[SYS_nice].rules[0] = &sandbox_deny;

  syscall_rules[SYS_execve].rules[0] = &execve_filter;
  //  syscall_rules[SYS_execve].rules[1] = &execve_logger;
  syscall_rules[SYS_execve].decoder = &execve_decoder;
  
  /**
   * Group and user related functions
   */
  syscall_rules[SYS_chroot].rules[0] = &sandbox_deny;
  syscall_rules[SYS_fchmod].rules[0] = &sandbox_deny;
  syscall_rules[SYS_fchown].rules[0] = &sandbox_deny;
  syscall_rules[SYS_chown].rules[0] = &sandbox_deny;
  syscall_rules[SYS_chown32].rules[0] = &sandbox_deny;
  syscall_rules[SYS_getegid].rules[0] = &sandbox_deny;
  syscall_rules[SYS_geteuid].rules[0] = &sandbox_deny;
  syscall_rules[SYS_getgid].rules[0] = &sandbox_deny;
  syscall_rules[SYS_getgroups].rules[0] = &sandbox_deny;
  syscall_rules[SYS_getpgrp].rules[0] = &sandbox_deny;
  syscall_rules[SYS_getuid].rules[0] = &sandbox_deny;
  syscall_rules[SYS_getuid32].rules[0] = &sandbox_deny;
  syscall_rules[SYS_lchown].rules[0] = &sandbox_deny;
  syscall_rules[SYS_lchown32].rules[0] = &sandbox_deny;
  syscall_rules[SYS_setgid].rules[0] = &sandbox_deny;
  syscall_rules[SYS_setgroups].rules[0] = &sandbox_deny;
  syscall_rules[SYS_setregid].rules[0] = &sandbox_deny;
  syscall_rules[SYS_setreuid].rules[0] = &sandbox_deny;
  syscall_rules[SYS_setsid].rules[0] = &sandbox_deny;
  syscall_rules[SYS_setpgid].rules[0] = &sandbox_deny;
  syscall_rules[SYS_setuid].rules[0] = &sandbox_deny;
  syscall_rules[SYS_setuid32].rules[0] = &sandbox_deny;
  syscall_rules[SYS_getegid32].rules[0] = &sandbox_deny;
  syscall_rules[SYS_getegid].rules[0] = &sandbox_deny;
  syscall_rules[SYS_getgid32].rules[0] = &sandbox_deny;
  syscall_rules[SYS_getgid].rules[0] = &sandbox_deny;
  syscall_rules[SYS_getgroups32].rules[0] = &sandbox_deny;
  syscall_rules[SYS_getgroups].rules[0] = &sandbox_deny;
  syscall_rules[SYS_getpgid].rules[0] = &sandbox_deny;
  syscall_rules[SYS_getresgid].rules[0] = &sandbox_deny;
  syscall_rules[SYS_getresgid32].rules[0] = &sandbox_deny;
  syscall_rules[SYS_getresuid].rules[0] = &sandbox_deny;
  syscall_rules[SYS_getresuid32].rules[0] = &sandbox_deny;
  syscall_rules[SYS_setfsgid].rules[0] = &sandbox_deny;
  syscall_rules[SYS_setfsgid32].rules[0] = &sandbox_deny;
  syscall_rules[SYS_setfsuid].rules[0] = &sandbox_deny;
  syscall_rules[SYS_setfsuid32].rules[0] = &sandbox_deny;
  syscall_rules[SYS_setregid].rules[0] = &sandbox_deny;
  syscall_rules[SYS_setregid32].rules[0] = &sandbox_deny;
  syscall_rules[SYS_setresuid].rules[0] = &sandbox_deny;
  syscall_rules[SYS_setresuid32].rules[0] = &sandbox_deny;
  syscall_rules[SYS_setresgid].rules[0] = &sandbox_deny;
  syscall_rules[SYS_setresgid32].rules[0] = &sandbox_deny;
  syscall_rules[SYS_setresuid].rules[0] = &sandbox_deny;
  syscall_rules[SYS_setresuid32].rules[0] = &sandbox_deny;

  /**
   * Socket Stuff
   */
  syscall_rules[SYS_socketcall].rules[0] = &socket_filter;


  /**
   * Signals rules : is allowed to manage, send signals to subprocess
   *    You can't change info on SIGTRAP signal
   *  see fork_functions.h for more info
   */
  syscall_rules[SYS_sigaltstack].rules[0] = &sandbox_deny;
  syscall_rules[SYS_signal].rules[0] = &signals_filter;
  syscall_rules[SYS_sigaction].rules[0] = &signals_filter;
  syscall_rules[SYS_sigaction].decoder = &sigaction_decoder;
  syscall_rules[SYS_rt_sigaction].rules[0] = &signals_filter;
  syscall_rules[SYS_rt_sigaction].decoder = &sigaction_decoder;
    syscall_rules[SYS_rt_sigprocmask].rules[0] = &sigset_filter;
  syscall_rules[SYS_rt_sigprocmask].decoder = &sigprocmask_decoder;
  syscall_rules[SYS_sigprocmask].rules[0] = &sigset_filter;
  syscall_rules[SYS_sigprocmask].decoder = &sigprocmask_decoder;
  syscall_rules[SYS_rt_sigpending].rules[0] = &sigset_filter;
  syscall_rules[SYS_rt_sigpending].decoder = &sigset_decoder;
  syscall_rules[SYS_sigpending].rules[0] = &sigset_filter;
  syscall_rules[SYS_sigpending].decoder = &sigset_decoder;
  syscall_rules[SYS_rt_sigtimedwait].rules[0] = &sigset_filter;
  syscall_rules[SYS_rt_sigtimedwait].decoder = &sigset_decoder;
  syscall_rules[SYS_kill].rules[0] = &kill_attitude;
#ifdef SYS_tgkill
  syscall_rules[SYS_tgkill].rules[0] = &sandbox_deny;
#endif
#ifdef SYS_tkill
  syscall_rules[SYS_tkill].rules[0] = &sandbox_deny;
#endif
  /**
   *  File stuffs : you can read, write and execute things only your current 
   * dirrectory and subs and read in few allowed part of the file system 
   * (like /usr/lib)
   */
  syscall_rules[SYS_mkdir].rules[0] = &file_filter;
  syscall_rules[SYS_mkdir].decoder = &rw_file_decoder;
  syscall_rules[SYS_rmdir].rules[0] = &file_filter;
  syscall_rules[SYS_rmdir].decoder = &rw_file_decoder;
  //  syscall_rules[SYS_open].rules[1] = &file_filter;
  syscall_rules[SYS_open].rules[0] = &file_rewriter_filter;
  syscall_rules[SYS_open].rules[1] = &file_rewriter_filter;
  syscall_rules[SYS_open].decoder = &rw_file_decoder;
  syscall_rules[SYS_creat].rules[0] = &file_filter;
  syscall_rules[SYS_creat].decoder = &rw_file_decoder;
  syscall_rules[SYS_link].rules[0] = &file_filter;
  syscall_rules[SYS_link].decoder = &link_file_decoder;
  syscall_rules[SYS_symlink].rules[0] = &file_filter;
  syscall_rules[SYS_symlink].decoder = &link_file_decoder;
  syscall_rules[SYS_chmod].rules[0] = &file_filter;
  syscall_rules[SYS_chmod].decoder = &r_file_decoder;
  syscall_rules[SYS_stat].rules[1] = &file_rewriter;
  syscall_rules[SYS_stat].decoder = &r_file_decoder;
  syscall_rules[SYS_stat64].rules[1] = &file_rewriter;
  syscall_rules[SYS_stat64].decoder = &r_file_decoder;
  syscall_rules[SYS_lstat].rules[0] = &file_filter;
  syscall_rules[SYS_lstat].decoder = &r_file_decoder;
  syscall_rules[SYS_oldlstat].rules[0] = &file_filter;
  syscall_rules[SYS_oldlstat].decoder = &r_file_decoder;
  syscall_rules[SYS_oldstat].rules[0] = &file_filter;
  syscall_rules[SYS_oldstat].decoder = &r_file_decoder;
  syscall_rules[SYS_lstat64].rules[0] = &file_filter;
  syscall_rules[SYS_lstat64].decoder = &r_file_decoder;
  syscall_rules[SYS_lstat].rules[0] = &file_filter;
  syscall_rules[SYS_lstat].decoder = &r_file_decoder;
  syscall_rules[SYS_unlink].rules[0] = &file_filter;
  syscall_rules[SYS_unlink].decoder = &r_file_decoder;
  syscall_rules[SYS_chdir].rules[0] = &file_filter;
  syscall_rules[SYS_chdir].decoder = &r_file_decoder;
  syscall_rules[SYS_truncate].rules[0] = &file_filter;
  syscall_rules[SYS_truncate].decoder = &rw_file_decoder;
  syscall_rules[SYS_truncate64].rules[0] = &file_filter;
  syscall_rules[SYS_truncate64].decoder = &rw_file_decoder;
  syscall_rules[SYS_utime].rules[0] = &file_filter;
  syscall_rules[SYS_utime].decoder = &rw_file_decoder;
#ifdef SYS_utimes
  syscall_rules[SYS_utimes].rules[0] = &file_filter;
  syscall_rules[SYS_utimes].decoder = &rw_file_decoder;
#endif//SYS_utimes
  syscall_rules[SYS_access].rules[1] = &access_rewriter; 
  syscall_rules[SYS_access].decoder = &r_file_decoder;
  //syscall_rules[SYS_access].rules[0] = &file_logger;
  //syscall_rules[SYS_access].rules[0] = &file_filter;
  return;
}
